#include "ObjectReader.h"
#include "ObjectRelationalReadDataBuffer.h"
#include "MappingManager.h"
#include "IContainerHandler.h"
#include "DataElement.h"
#include "RelationalStorageServiceNames.h"
#include "RelationalStorageServiceException.h"
#include "ObjectRelationalAccess/ObjectRelationalMapping.h"
#include "ObjectRelationalAccess/ObjectRelationalMappingElement.h"
#include "RelationalAccess/ITable.h"
#include "RelationalAccess/IQuery.h"
#include "RelationalAccess/ICursor.h"
#include "CoralBase/Attribute.h"
#include "CoralBase/AttributeList.h"
#include "CoralBase/AttributeListSpecification.h"
#include "CoralBase/AttributeSpecification.h"
#include "CoralBase/Blob.h"
#include "StorageSvc/DataCallBack.h"
#include "StorageSvc/DbDataHandlerGuard.h"
#include "POOLCore/Ptr.h"
#include "POOLCore/PContainer.h"
#include "POOLCore/Reference.h"
#include "POOLCore/Token.h"
#include "Reflex/Reflex.h"
#include <memory>
#include <sstream>

namespace pool {
  namespace RelationalStorageService {
    union U_Primitives {
      bool b;
      char c;
      unsigned char uc;
      short s;
      unsigned short us;
      int i;
      unsigned int ui;
      long l;
      unsigned long ul;
      float f;
      double d;
      long double ld;
    };
    
    class RelationalPtrLoader: virtual public pool::IPtrLoader {
      public:
        RelationalPtrLoader(ObjectReader& reader,
                            ObjectRelationalReadDataBuffer& buffer,
                            void* objectData,
                            int oid,
                            bool anyObject,
                            pool::DataCallBack& callback ):
          IPtrLoader(),
          m_isValid(true),
          m_reader(reader),
          m_buffer(buffer),
          m_objectData(objectData),
          m_oid(oid),
          m_anyObject(anyObject),
          m_callback(callback){}

        virtual ~RelationalPtrLoader(){}
        
        void* load() const {
          void* ret = 0;
          if(m_isValid) ret = m_reader.loadPoolPointer(m_buffer,m_oid,m_anyObject,m_callback);
          return ret;
        }

        void notify() const {
          if(m_isValid) {
            m_reader.eraseLoader(this);
          }
        }

        void invalidate() {
          m_isValid = false;
        }

        bool isValid() const {
          return m_isValid;
        }

      private:
        bool m_isValid;
        ObjectReader& m_reader;
        ObjectRelationalReadDataBuffer& m_buffer;
        void* m_objectData;
        int m_oid;
        bool m_anyObject;
        pool::DataCallBack& m_callback;
    };


    class RelationalContainerLoader: public pool::IContainerLoader {

      public:

        // constructor
        RelationalContainerLoader(ObjectReader& reader,
                                  ObjectRelationalReadDataBuffer& buffer,
                                  const std::vector<int>& identity,
                                  pool::DataCallBack& callback):
          IContainerLoader(),
          m_isValid(true),
          m_reader(reader),
          m_buffer(buffer),
          m_identity(identity),
          m_callback(callback){
        }
        

        // destructor
        virtual ~RelationalContainerLoader(){
        }

      public:

        // triggers the data loading
        bool load(void* address) const {
          bool ret = false;
          if(m_isValid) {
            Selection empty;
            ret = m_reader.loadPoolContainer(m_identity, empty, m_buffer,address,m_callback);
          }
          return ret;
        }
        

        bool loadSelection(const Selection& selection, void* address) const {
          bool ret = false;
          if(m_isValid) {
            ret = m_reader.loadPoolContainer(m_identity, selection, m_buffer,address,m_callback);
          }
          return ret;          
        }
        

        size_t getSelectionCount( const Selection& selection ) const {
          size_t ret = 0;
          if(m_isValid) {
            ret = m_reader.makeCountQuery(m_identity, selection, m_buffer);
          }
          return ret;          
        }

        // notify the underlying storage system that the embedded object has been destructed.
        void notify() const {
          if(m_isValid) {
            m_reader.eraseLoader(this);
          }
        }

        // invalidates the current loader. Called by the underlying service at his destruction time.
        void invalidate(){
          m_isValid = false;
        }
        

        // queries the validity of the current relation with the underlying storage system
        bool isValid() const{
          return m_isValid;
        }
        
      private:
        bool m_isValid;
        ObjectReader& m_reader;
        ObjectRelationalReadDataBuffer& m_buffer;
        std::vector<int> m_identity;
        pool::DataCallBack& m_callback;

    };
  
  }
}


pool::RelationalStorageService::ObjectReader::ObjectReader( coral::ISchema& schema,
                                                            const Reflex::Type& dictionaryEntry,
                                                            MappingManager& mappingManager,
                                                            pool::DbDatabase* databaseHandle ):
  DbIOHandler(databaseHandle),
  m_schema(schema),
  m_dataBuffer( 0 ),
  m_refDict( Reflex::Type::ByTypeInfo( typeid(pool::Reference) ) ),
  m_mappingManager(mappingManager),
  m_poolPointerLoaders(),
  m_poolContainerLoaders()
{
  m_dataBuffer = new pool::RelationalStorageService::ObjectRelationalReadDataBuffer(m_schema, m_context);
  const ObjectRelationalMappingElement& mappingElement = m_mappingManager.mappingForClass( dictionaryEntry ).element();
  m_dataBuffer->build( dictionaryEntry, mappingElement );
}

pool::RelationalStorageService::ObjectReader::~ObjectReader()
{
  delete m_dataBuffer;
  for(std::set<RelationalPtrLoader*>::iterator iPtr=m_poolPointerLoaders.begin();
      iPtr!=m_poolPointerLoaders.end();
      ++iPtr){
    (*iPtr)->invalidate();
  }
  m_poolPointerLoaders.clear();
  for(std::set<RelationalContainerLoader*>::iterator iPtr=m_poolContainerLoaders.begin();
      iPtr!=m_poolContainerLoaders.end();
      ++iPtr){
    (*iPtr)->invalidate();
  }
  m_poolContainerLoaders.clear();
}

bool
pool::RelationalStorageService::ObjectReader::readObject( void* objectData,
                                                          int oid,
                                                          bool anyObject,
                                                          pool::DataCallBack& callback )
{
  std::auto_ptr< coral::IQuery > query( makeQuery(*m_dataBuffer,oid,anyObject) );
  coral::ICursor& cursor = query->execute();

  if ( cursor.next() ) {
    
    readObjectData( cursor,*m_dataBuffer,objectData,oid,callback );
    processDependentBuffers( *m_dataBuffer, objectData, oid, anyObject, callback);
    
    return true;
  }
  else {
    return false;
  }
}

void*
pool::RelationalStorageService::ObjectReader::loadPoolPointer( ObjectRelationalReadDataBuffer& buffer,
                                                               int oid,
                                                               bool anyObject,
                                                               pool::DataCallBack& callback ){
  void* ptrData = 0;
  std::auto_ptr< coral::IQuery > query(makeQuery( buffer,oid,anyObject));
  coral::ICursor& cursor = query->execute();
  if ( cursor.next() ) {
    Reflex::Type ptrType;
    if( buffer.mapping().elementType()==pool::ObjectRelationalMappingElement::PoolPointer ){
      ptrType = buffer.dictionary().TemplateArgumentAt(0);
    } else if ( buffer.mapping().elementType()==pool::ObjectRelationalMappingElement::DependentObject ){
      ptrType = buffer.dictionary();
    }
    if(!ptrType){
        throw pool::RelationalStorageService::RelationalStorageServiceException(
          "Could not resolve the pool pointer class.",
          "ObjectReader::loadPoolPointer" );
    }
    
    ptrData = ptrType.Allocate();
    ptrType.Construct(ROOT::Reflex::Type(0,0),std::vector<void*>(),ptrData);
    readObjectData( cursor,buffer,ptrData,oid,callback );
    processDependentBuffers( buffer, ptrData, oid, anyObject, callback);
  }
  return ptrData;
}

bool pool::RelationalStorageService::ObjectReader::loadPoolContainer( const std::vector<int>& identity,
                                                                      const pool::Selection& selection,
                                                                      ObjectRelationalReadDataBuffer& buffer,
                                                                      void* address,
                                                                      pool::DataCallBack& callback ){
  std::auto_ptr< coral::IQuery > query( buffer.table().newQuery() );
  const coral::AttributeListSpecification& outputSpecification = buffer.dataSpecification();
  for (coral::AttributeListSpecification::const_iterator iSp=outputSpecification.begin(); iSp != outputSpecification.end(); ++iSp ) {
    query->addToOutputList( iSp->name() );
    query->defineOutputType( iSp->name(),iSp->typeName());
  }

  coral::AttributeListSpecification::const_iterator iSpec = outputSpecification.begin();
  coral::AttributeList whereData(buffer.whereDataForId());
  coral::AttributeList::iterator iData = whereData.begin();

  if( buffer.mapping().isDependentTree() ) {
    // skip the refid column for dependent objects
    ++iSpec;
    // set the identity for the refid
    iData->data<int>() = (buffer.parentMetadata()->operator[](buffer.parentMetadataOffset())).second;
    ++iData;
  }
  
  query->addToOrderList( iSpec->name() ); // Order by pos.

  // Fill-in the identities.  
  for ( std::vector<int>::const_iterator id = identity.begin();
        id != identity.end(); ++id ) {
    iData->data<int>() = *id;
    ++iData;
  }

  std::stringstream condition;
  condition << buffer.whereClauseForId();
  if(!selection.isEmpty()){
    condition << setQueryCondition(selection, buffer.mapping(), whereData);
  }
  
  // Set the condition
  query->setCondition( condition.str(), whereData );

  // Prefetch rows.
  query->setRowCacheSize( 100 ); // We should better define this value !!!

  // Create the identity for the sub-elements
  std::vector<int> identityForSubElements( 1 + identity.size(), 0 );

  // Check if the container is an associative one
  Reflex::Type dictionary = buffer.dictionary();
  Reflex::Type iteratorDereferenceReturnType = buffer.containerHandler().iteratorReturnType();
  Reflex::Member firstMember = iteratorDereferenceReturnType.MemberByName( "first" );
  if ( ! firstMember ) {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Could not retrieve the data member \"first\" for the class \"" +
                                                                             iteratorDereferenceReturnType.Name(Reflex::SCOPED|Reflex::FINAL) + "\"",
                                                                             "ObjectReader::loadPoolContainer" );
  }
  Reflex::Member secondMember = iteratorDereferenceReturnType.MemberByName( "second" );
  if ( ! secondMember ) {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Could not retrieve the data member \"second\" for the class \"" +
                                                                             iteratorDereferenceReturnType.Name(Reflex::SCOPED|Reflex::FINAL) + "\"",
                                                                             "ObjectReader::loadPoolContainer" );
  }
  
  // Issue the query...
  coral::ICursor& cursor = query->execute();

  buffer.containerHandler().clear( address );
  //unsigned int i=0;
  while ( cursor.next() ) {

    const coral::AttributeList& row = cursor.currentRow();
    
    // Create a new element for the array
    void* objectData = iteratorDereferenceReturnType.Construct().Address();    
    void* positionData = static_cast< char* >( objectData ) + firstMember.Offset();
    void* containerData = static_cast< char* >( objectData ) + secondMember.Offset();
    
    // Get the Identity
    coral::AttributeList::const_iterator iAttribute = row.begin();
    if( buffer.mapping().isDependentTree() ) {
      // skip the refid column
      ++iAttribute;
    }
    std::vector<int>::iterator idout = identityForSubElements.begin();
    int pos = iAttribute->data< int >();
    *idout = pos;
    ++idout;
    ++iAttribute;
    *(size_t*)positionData = (size_t)(pos);
    for ( ; idout != identityForSubElements.end(); ++idout ) {
      *idout = iAttribute->data< int >();
      ++iAttribute;
    }

    // Copy the data
    std::vector<DataElement*>::const_iterator iDataElement = buffer.dataElements().begin();

    while ( iAttribute != row.end() ) {
      DataElement& dataElement = *(*iDataElement);
      void* dataElementAddress = dataElement.address( containerData );
      if( dataElement.hasMetadata()){
        buffer.metadata()[ dataElement.metadataIndex() ].first = iAttribute->data<std::string>();
        ++iAttribute;
        buffer.metadata()[ dataElement.metadataIndex() ].second = iAttribute->data<int>();
      } else {
        if( dataElement.isRef() ){
          pool::Token::OID_t oid;
          oid.first = iAttribute->data<int>();
          ++iAttribute;
          ++iDataElement;
          oid.second = iAttribute->data<int>();
          // Call here the ref handler to get the oid...
          pool::DbDataHandlerGuard handlerGuard( &callback, m_databaseContext );
          callback.notify( pool::DataCallBack::GET, pool::DataCallBack::REFERENCE, m_refDict, dataElementAddress, &oid );
        } else {
          if( iAttribute->specification().type() == typeid(coral::Blob) ){
            buffer.readBlob( iAttribute->specification().name(), iAttribute->data<coral::Blob>(), dataElementAddress );
          } else {
            iAttribute->copyValueToAddress( dataElementAddress );
          }          
        }
      }
      ++iAttribute;
      ++iDataElement;
    }

    // Loop over the arrays now...
    for ( pool::RelationalStorageService::ObjectRelationalReadDataBuffer::iterator iBuffer = buffer.begin();
          iBuffer != buffer.end(); ++iBuffer ) {
      if ( ! this->readArray( identityForSubElements, **iBuffer,
                              (*iBuffer)->parentDataElement().address( containerData ),
                              callback ) ) return false;
    }
    // Loop over the arrays now...
    for ( std::vector<ObjectRelationalReadDataBuffer*>::const_iterator iBuffer = buffer.poolContainerBuffers().begin();
          iBuffer != buffer.poolContainerBuffers().end(); ++iBuffer ) {
      if ( ! this->readPoolContainer( identityForSubElements, **iBuffer,
                                      (*iBuffer)->parentDataElement().address( containerData ),
                                      callback ) ) return false;
    }
    // Loop over the pool pointers...
    for ( std::vector<ObjectRelationalReadDataBuffer*>::const_iterator iBuffer = buffer.poolPointerBuffers().begin();
          iBuffer != buffer.poolPointerBuffers().end(); ++iBuffer ) {
      if ( ! this->readPoolPointer( *(*iBuffer),
                                    (*iBuffer)->parentDataElement().address( containerData ),
                                    identityForSubElements.back(),
                                    false,
                                    callback ) ) return false;
    }
    // Loop over the pool polymorhic pointers...
    for ( std::vector<ObjectRelationalReadDataBuffer*>::const_iterator iBuffer = buffer.poolPolymorphicPointerBuffers().begin();
          iBuffer != buffer.poolPolymorphicPointerBuffers().end(); ++iBuffer ) {
      if ( ! this->readPoolPolymorphicPointer( *(*iBuffer),
                                               (*iBuffer)->parentDataElement().address( containerData ),
                                               identityForSubElements.back(),
                                               false,
                                               callback ) ) return false;
    }

    size_t prevSize = buffer.containerHandler().size( address );
    buffer.containerHandler().appendNewElement( address, objectData );
    bool inserted = buffer.containerHandler().size( address )>prevSize;
    if ( ! ( iteratorDereferenceReturnType.IsFundamental() ) ) {
      iteratorDereferenceReturnType.Destruct( objectData );
    }
    if ( !inserted ) return false;
  }

  buffer.containerHandler().finalize( address );
  
  return true;  
}

size_t
pool::RelationalStorageService::ObjectReader::makeCountQuery( const std::vector<int>& identity,
                                                              const Selection& selection,
                                                              ObjectRelationalReadDataBuffer& buffer ){
  std::auto_ptr< coral::IQuery > query( buffer.table().newQuery() );
  std::string countColumn("COUNT(*)");
  query->addToOutputList( countColumn );
  query->defineOutputType( countColumn, coral::AttributeSpecification::typeNameForType<int>() );

  coral::AttributeList whereData(buffer.whereDataForId());
  coral::AttributeList::iterator iData = whereData.begin();

  if( buffer.mapping().isDependentTree() ) {
    // set the identity for the refid
    iData->data<int>() = (buffer.parentMetadata()->operator[](buffer.parentMetadataOffset())).second;
    ++iData;
  }
  

  // Fill-in the identities.  
  for ( std::vector<int>::const_iterator id = identity.begin();
        id != identity.end(); ++id ) {
    iData->data<int>() = *id;
    ++iData;
  }

  std::stringstream condition;
  condition << buffer.whereClauseForId();
  if(!selection.isEmpty()){
    condition << setQueryCondition(selection, buffer.mapping(), whereData);
  }
  
  // Set the condition
  query->setCondition( condition.str(), whereData );

  // Issue the query...
  coral::ICursor& cursor = query->execute();

  size_t result = 0;
  while ( cursor.next() ) {

    const coral::AttributeList& row = cursor.currentRow();
    result = row[countColumn].data<int>();
  }
  return result;
}

bool pool::RelationalStorageService::ObjectReader::eraseLoader( const RelationalPtrLoader* loader ){
  return m_poolPointerLoaders.erase(const_cast<RelationalPtrLoader*>(loader));
}

bool pool::RelationalStorageService::ObjectReader::eraseLoader( const RelationalContainerLoader* loader ){
  return m_poolContainerLoaders.erase(const_cast<RelationalContainerLoader*>(loader));
}

bool
pool::RelationalStorageService::ObjectReader::readArray( const std::vector<int>& identity,
                                                         ObjectRelationalReadDataBuffer& buffer,
                                                         void* address,
                                                         pool::DataCallBack& callback )
{
  std::auto_ptr< coral::IQuery > query( buffer.table().newQuery() );
  const coral::AttributeListSpecification& outputSpecification = buffer.dataSpecification();
  for (coral::AttributeListSpecification::const_iterator iSp=outputSpecification.begin(); iSp != outputSpecification.end(); ++iSp ) {
    query->addToOutputList( iSp->name() );
    query->defineOutputType( iSp->name(),iSp->typeName());
  }

  coral::AttributeListSpecification::const_iterator iSpec = outputSpecification.begin();
  coral::AttributeList& whereData = buffer.whereDataForId();
  coral::AttributeList::iterator iData = whereData.begin();

  if( buffer.mapping().isDependentTree() ) {
    // skip the refid column for dependent objects
    ++iSpec;
    // set the identity for the refid
    iData->data<int>() = (buffer.parentMetadata()->operator[](buffer.parentMetadataOffset())).second;
    ++iData;
  }
  
  query->addToOrderList( iSpec->name() ); // Order by pos.

  // Fill-in the identities.  
  for ( std::vector<int>::const_iterator id = identity.begin();
        id != identity.end(); ++id ) {
    iData->data<int>() = *id;
    ++iData;
  }

  // Set the condition
  query->setCondition( buffer.whereClauseForId(), whereData );

  // Prefetch rows.
  query->setRowCacheSize( 100 ); // We should better define this value !!!

  // Create the identity for the sub-elements
  std::vector<int> identityForSubElements( 1 + identity.size(), 0 );

  // Check if the container is an associative one
  Reflex::Type keyType;
  Reflex::Type dictionary = buffer.dictionary();
  Reflex::Type iteratorDereferenceReturnType = buffer.containerHandler().iteratorReturnType();
  pool::RelationalStorageService::U_Primitives primitiveStub;
  int keyItems = buffer.keyItems();
  
  Reflex::Member firstMember;
  if ( keyItems > 0 ) {
    keyType = dictionary.TemplateArgumentAt(0);
    if(keyType){
      firstMember = iteratorDereferenceReturnType.MemberByName( "first" );
      if ( ! firstMember ) {
        throw pool::RelationalStorageService::RelationalStorageServiceException(
          "Could not retrieve the data member \"first\" of the class \"" +
          iteratorDereferenceReturnType.Name(Reflex::SCOPED|Reflex::FINAL) + "\"",
          "ObjectReader::readArray" );
      }
    }
  }
  
  // Issue the query...
  coral::ICursor& cursor = query->execute();

  bool isArray = dictionary.IsArray();
  bool isElementFundamental = iteratorDereferenceReturnType.IsFundamental();

  size_t arraySize = dictionary.ArrayLength();
  
  buffer.containerHandler().clear( address );
  unsigned int i=0;
  while ( cursor.next() ) {

    const coral::AttributeList& row = cursor.currentRow();
    
    // Get the Identity
    coral::AttributeList::const_iterator iAttribute = row.begin();
    if( buffer.mapping().isDependentTree() ) {
      // skip the refid column
      ++iAttribute;
    }
    int arrayIndex = iAttribute->data< int >();
    for ( std::vector<int>::iterator idout = identityForSubElements.begin();
          idout != identityForSubElements.end(); ++idout ) {
      *idout = iAttribute->data< int >();
      ++iAttribute;
    }

    // Create a new element for the array
    void* objectData = 0;
    if( isArray ){
      if(arrayIndex < (int)arraySize){
        // the memory has been allocated already!
        objectData = static_cast<char*>(address)+arrayIndex*iteratorDereferenceReturnType.SizeOf();
        if(!isElementFundamental){
          // in this case the initialization is required: use default constructor...
          iteratorDereferenceReturnType.Construct(Reflex::Type(0,0),std::vector< void* >(),objectData);
        }
      } else {
        // not sure what to do in this case... exception or ignore???
        continue;
      }
    } else {
      if(isElementFundamental){
        objectData = &primitiveStub;
      } else {
        objectData = iteratorDereferenceReturnType.Construct().Address();
      }
    }

    void* componentData = objectData;
    void* keyData = 0;

    // Copy the data
    std::vector<DataElement*>::const_iterator iDataElement = buffer.dataElements().begin();

    if ( keyType ) { // treat the key object first
      keyData = static_cast< char* >( objectData ) + firstMember.Offset();
      for ( int i = 0; i < keyItems; ++i ) {
        DataElement& dataElement = *(*iDataElement);
        void* dataElementAddress = dataElement.address( keyData );
        if( dataElement.hasMetadata()){
          buffer.metadata()[ dataElement.metadataIndex() ].first = iAttribute->data<std::string>();
          ++iAttribute;
          buffer.metadata()[ dataElement.metadataIndex() ].second = iAttribute->data<int>();
        } else {
          if( dataElement.isRef() ){
            pool::Token::OID_t oid;
            oid.first = iAttribute->data<int>();
            ++iAttribute;
            ++iDataElement;
            oid.second = iAttribute->data<int>();
            // Call here the ref handler to get the oid...
            pool::DbDataHandlerGuard handlerGuard( &callback, m_databaseContext );
            callback.notify( pool::DataCallBack::GET, pool::DataCallBack::REFERENCE, m_refDict, dataElementAddress, &oid );
          } else {
            if( iAttribute->specification().type() == typeid(coral::Blob) ){
              buffer.readBlob( iAttribute->specification().name(), iAttribute->data<coral::Blob>(), dataElementAddress );
            } else {
              iAttribute->copyValueToAddress( dataElementAddress );
            }            
          }
        }
        ++iAttribute;
        ++iDataElement;
      }
      const Reflex::Member secondMember = iteratorDereferenceReturnType.MemberByName( "second" );
      if ( ! secondMember ) {
        throw pool::RelationalStorageService::RelationalStorageServiceException( "Could not retrieve the data member \"second\" of the class \"" +
                                                                                   iteratorDereferenceReturnType.Name(Reflex::SCOPED|Reflex::FINAL) + "\"",
                                                                                 "ObjectReader::readArray" );
      }
      componentData = static_cast< char* >( objectData ) + secondMember.Offset();
    }

    while ( iAttribute != row.end() ) {
      DataElement& dataElement = *(*iDataElement);
      void* dataElementAddress = dataElement.address( componentData );
      if( dataElement.hasMetadata()){
        buffer.metadata()[ dataElement.metadataIndex() ].first = iAttribute->data<std::string>();
        ++iAttribute;
        buffer.metadata()[ dataElement.metadataIndex() ].second = iAttribute->data<int>();
      } else {
        if( dataElement.isRef() ){
          pool::Token::OID_t oid;
          oid.first = iAttribute->data<int>();
          ++iAttribute;
          ++iDataElement;
          oid.second = iAttribute->data<int>();
          // Call here the ref handler to get the oid...
          pool::DbDataHandlerGuard handlerGuard( &callback, m_databaseContext );
          callback.notify( pool::DataCallBack::GET, pool::DataCallBack::REFERENCE, m_refDict, dataElementAddress, &oid );
        } else {
          if( iAttribute->specification().type() == typeid(coral::Blob) ){
            buffer.readBlob( iAttribute->specification().name(), iAttribute->data<coral::Blob>(), dataElementAddress );
          } else {
            iAttribute->copyValueToAddress( dataElementAddress );
          }          
        }
      }
      ++iAttribute;
      ++iDataElement;
    }

    // Loop over the arrays now...
    for ( pool::RelationalStorageService::ObjectRelationalReadDataBuffer::iterator iBuffer = buffer.begin();
          iBuffer != buffer.end(); ++iBuffer ) {
      void* theAddress = componentData;
      if ( (*iBuffer)->isKeyData() ) {
        theAddress = keyData;
      }
      if ( ! this->readArray( identityForSubElements, **iBuffer,
                              (*iBuffer)->parentDataElement().address( theAddress ),
                              callback ) ) return false;
    }
    // Loop over the pool containers now...
    for ( std::vector<ObjectRelationalReadDataBuffer*>::const_iterator iBuffer = buffer.poolContainerBuffers().begin();
          iBuffer != buffer.poolContainerBuffers().end(); ++iBuffer ) {
      void* theAddress = componentData;
      if ( (*iBuffer)->isKeyData() ) {
        theAddress = keyData;
      }
      if ( ! this->readPoolContainer( identityForSubElements, **iBuffer,
                                      (*iBuffer)->parentDataElement().address( theAddress ),
                                      callback ) ) return false;
    }
    // Loop over the pool pointers...
    for ( std::vector<ObjectRelationalReadDataBuffer*>::const_iterator iBuffer = buffer.poolPointerBuffers().begin();
          iBuffer != buffer.poolPointerBuffers().end(); ++iBuffer ) {
      void* theAddress = componentData;
      if ( (*iBuffer)->isKeyData() ) {
        theAddress = keyData;
      }
      if ( ! this->readPoolPointer( *(*iBuffer),
                                    (*iBuffer)->parentDataElement().address( theAddress ),
                                    identityForSubElements.back(),
                                    false,
                                    callback ) ) return false;
    }
    // Loop over the pool polymorhic pointers...
    for ( std::vector<ObjectRelationalReadDataBuffer*>::const_iterator iBuffer = buffer.poolPolymorphicPointerBuffers().begin();
          iBuffer != buffer.poolPolymorphicPointerBuffers().end(); ++iBuffer ) {
      void* theAddress = componentData;
      if ( (*iBuffer)->isKeyData() ) {
        theAddress = keyData;
      }
      if ( ! this->readPoolPolymorphicPointer( *(*iBuffer),
                                               (*iBuffer)->parentDataElement().address( theAddress ),
                                               identityForSubElements.back(),
                                               false,
                                               callback ) ) return false;
    }

    if(!isArray) {
      size_t prevSize = buffer.containerHandler().size( address );
      buffer.containerHandler().appendNewElement( address, objectData );
      bool inserted = buffer.containerHandler().size( address )>prevSize;
      if ( ! ( iteratorDereferenceReturnType.IsFundamental() ) ) {
        iteratorDereferenceReturnType.Destruct( objectData );
      }
      if ( !inserted ) return false;
    }
    ++i;
  }

  buffer.containerHandler().finalize( address );
  
  return true;
}

bool
pool::RelationalStorageService::ObjectReader::readPoolContainer( const std::vector<int>& identity,
                                                                 ObjectRelationalReadDataBuffer& buffer,
                                                                 void* address,
                                                                 pool::DataCallBack& callback ){
  Reflex::Type poolContainerType = buffer.dictionary();
  poolContainerType.UpdateMembers();
  Reflex::Object poolContainerObj(poolContainerType,address);

  RelationalContainerLoader* loader = new RelationalContainerLoader(*this,buffer,identity,callback);
  m_poolContainerLoaders.insert(loader);
  
#if ROOT_VERSION_CODE < ROOT_VERSION(5,21,3)
  poolContainerObj.Invoke("installLoader",std::vector<void*>(1,loader));
#else
  poolContainerObj.Invoke("installLoader",0,std::vector<void*>(1,loader));
#endif
  return true;
}

bool
pool::RelationalStorageService::ObjectReader::readPoolPointer( ObjectRelationalReadDataBuffer& dataBuffer,
                                                               void* objectData,
                                                               int oid,
                                                               bool anyObject,
                                                               pool::DataCallBack& callback ){
  Reflex::Type poolPtrType = dataBuffer.dictionary();
  Reflex::Object poolPtrObj(poolPtrType,objectData);
  poolPtrObj.Invoke("reset");
  
  Reflex::Member ptrMember = poolPtrType.MemberByName("m_loader");
  RelationalPtrLoader* loader = new RelationalPtrLoader(*this,dataBuffer,objectData,oid,anyObject,callback);
  loader->addRef();
  *(pool::IPtrLoader**)(static_cast<char*>(objectData)+ptrMember.Offset())=loader;
  m_poolPointerLoaders.insert(loader);
  return true;
}

bool
pool::RelationalStorageService::ObjectReader::readPoolPolymorphicPointer( ObjectRelationalReadDataBuffer& dataBuffer,
                                                                          void* objectData,
                                                                          int oid,
                                                                          bool anyObject,
                                                                          pool::DataCallBack& callback ){
  const std::string& className = (*dataBuffer.parentMetadata())[dataBuffer.parentMetadataOffset()].first;
  if( className == RelationalStorageServiceNames::polyPtrNullLabel() ) return true;
  Reflex::Type ptrType = Reflex::Type::ByName(className);
  if( ptrType ){
    const pool::ObjectRelationalMappingElement& classMapping =  m_mappingManager.mappingForDependingClass( ptrType, m_dataBuffer->dictionary() );
    pool::RelationalStorageService::ObjectRelationalReadDataBuffer& newDataBuffer = dataBuffer.addDependency();
    newDataBuffer.build( ptrType, classMapping );
    Reflex::Type poolPtrType = dataBuffer.dictionary();
    Reflex::Object poolPtrObj(poolPtrType,objectData);
    poolPtrObj.Invoke("reset");

    Reflex::Member ptrMember = poolPtrType.MemberByName("m_loader");
    RelationalPtrLoader* loader = new RelationalPtrLoader(*this,newDataBuffer,objectData,oid,anyObject,callback);
    loader->addRef();
    *(pool::IPtrLoader**)(static_cast<char*>(objectData)+ptrMember.Offset())=loader;
    m_poolPointerLoaders.insert(loader);
    return true;
  }

  return false;
}

coral::IQuery*
pool::RelationalStorageService::ObjectReader::makeQuery( ObjectRelationalReadDataBuffer& dataBuffer,
                                                         int oid,
                                                         bool anyObject){
  coral::IQuery* query = dataBuffer.table().newQuery();
  const coral::AttributeListSpecification& outputSpecification = dataBuffer.dataSpecification();
  for ( coral::AttributeListSpecification::const_iterator iSpec = outputSpecification.begin();
        iSpec != outputSpecification.end(); ++iSpec ) {
    query->addToOutputList( iSpec->name() );
    query->defineOutputType( iSpec->name(),iSpec->typeName());
  }

  coral::AttributeList& oidData = dataBuffer.whereDataForId();
  std::stringstream whereClause;
  bool more = false;
  coral::AttributeList::iterator iAttr=oidData.begin();
  if(dataBuffer.mapping().isDependentTree()){
    whereClause << iAttr->specification().name();
    whereClause << "=";
    whereClause << " :" << iAttr->specification().name();
    iAttr->data<int>() = (dataBuffer.parentMetadata()->operator[](dataBuffer.parentMetadataOffset())).second;
    ++iAttr;
    more = true;
  } 
  for(; iAttr!=oidData.end();++iAttr){
    if(more) whereClause << " AND ";
    whereClause << iAttr->specification().name();
    iAttr->data< int >() = oid;
    if ( anyObject ) {
      whereClause << ">";
      query->addToOrderList( iAttr->specification().name() );
    } else whereClause << "=";
    whereClause << " :" << iAttr->specification().name();
    more = true;
  }
  if ( anyObject ) {
    query->limitReturnedRows( 1, 0 );
  }  
  query->setCondition( whereClause.str(), oidData );
  return query;
}

bool
pool::RelationalStorageService::ObjectReader::readObjectData( coral::ICursor& queryData,
                                                              ObjectRelationalReadDataBuffer& dataBuffer,
                                                              void* objectData,
                                                              int& oid,
                                                              pool::DataCallBack& callback ){
  const coral::AttributeList& row = queryData.currentRow();

  // Get the OID
  coral::AttributeList::const_iterator iAttribute = row.begin();
  if(dataBuffer.mapping().isDependentTree()){
    ++iAttribute; // in this case, ignoring the first column (ref_column)  
  }
  oid = iAttribute->data< int >();
  
  ++iAttribute;

  // Copy the data
  std::vector<DataElement*>::const_iterator iDataElement = dataBuffer.dataElements().begin();
  while ( iAttribute != row.end() ) {
    DataElement& dataElement = *(*iDataElement);
    void* dataElementAddress = dataElement.address( objectData );
    if( dataElement.hasMetadata()){
      dataBuffer.metadata()[ dataElement.metadataIndex() ].first = iAttribute->data<std::string>();
      ++iAttribute;
      dataBuffer.metadata()[ dataElement.metadataIndex() ].second = iAttribute->data<int>(); 
    } else {
      if( dataElement.isRef() ){
        pool::Token::OID_t toid;
        toid.first = iAttribute->data<int>();
        ++iAttribute;
        ++iDataElement;
        toid.second = iAttribute->data<int>();
        // Call here the ref handler to get the oid...
        pool::DbDataHandlerGuard handlerGuard( &callback, m_databaseContext );
        callback.notify( pool::DataCallBack::GET, pool::DataCallBack::REFERENCE, m_refDict, dataElementAddress, &toid );
      } else {
        if( iAttribute->specification().type() == typeid(coral::Blob) ){
          dataBuffer.readBlob( iAttribute->specification().name(), iAttribute->data<coral::Blob>(), dataElementAddress );
        } else {
          iAttribute->copyValueToAddress( dataElementAddress );
        }        
      }  
    }
    ++iAttribute;
    ++iDataElement;
  }
  return true;
}

bool
pool::RelationalStorageService::ObjectReader::processDependentBuffers( ObjectRelationalReadDataBuffer& dataBuffer,
                                                                       void* objectData,
                                                                       int oid,
                                                                       bool anyObject,
                                                                       pool::DataCallBack& callback ){
  // Loop over the arrays now...
  std::vector< int > identity( 1, oid );
  for ( pool::RelationalStorageService::ObjectRelationalReadDataBuffer::iterator iBuffer = dataBuffer.begin();
        iBuffer != dataBuffer.end(); ++iBuffer ) {
    //(*iBuffer)->containerHandler().clear((*iBuffer)->parentDataElement().address( objectData ));
    if ( ! this->readArray( identity, **iBuffer,
                            (*iBuffer)->parentDataElement().address( objectData ),
                            callback ) ) return false;
  }
  // Loop over the pool containers now...
  for ( std::vector<ObjectRelationalReadDataBuffer*>::const_iterator iBuffer = dataBuffer.poolContainerBuffers().begin();
        iBuffer != dataBuffer.poolContainerBuffers().end(); ++iBuffer ) {
    if ( ! this->readPoolContainer( identity, **iBuffer,
                                    (*iBuffer)->parentDataElement().address( objectData ),
                                    callback ) ) return false;
  }
  // Loop over the pool pointers now...
  for ( std::vector<ObjectRelationalReadDataBuffer*>::const_iterator iBuffer = dataBuffer.poolPointerBuffers().begin();
        iBuffer != dataBuffer.poolPointerBuffers().end(); ++iBuffer ) {
    if ( ! this->readPoolPointer( *(*iBuffer),
                                  (*iBuffer)->parentDataElement().address( objectData ),
                                  oid,
                                  anyObject,
                                  callback ) ) return false;
  }
  // Loop over the pool polymorphic pointers now...
  for ( std::vector<ObjectRelationalReadDataBuffer*>::const_iterator iBuffer = dataBuffer.poolPolymorphicPointerBuffers().begin();
        iBuffer != dataBuffer.poolPolymorphicPointerBuffers().end(); ++iBuffer ) {
    if ( ! this->readPoolPolymorphicPointer( *(*iBuffer),
                                             (*iBuffer)->parentDataElement().address( objectData ),
                                             oid,
                                             anyObject,
                                             callback ) ) return false;
  }
  return true;
}

std::string
pool::RelationalStorageService::ObjectReader::setQueryCondition(const pool::Selection& selection,
                                                                const ObjectRelationalMappingElement& mappingElement,
                                                                coral::AttributeList& conditionData){
  const std::vector<std::pair<std::string,std::string> >& theItems =  selection.items();
  std::stringstream cond;
  cond << " AND ";
  unsigned int i=0;
  for(std::vector<std::pair<std::string,std::string> >::const_iterator iItem = theItems.begin();
      iItem != theItems.end();
      ++iItem){
    if(i>0) cond << " AND ";
    std::string varName = pool::Selection::variableNameFromUniqueString(iItem->first);
    std::stringstream selColumn;
    std::string colName("");
    if(varName == pool::Selection::indexVariable()){
      size_t ix = 0;
      if(mappingElement.isDependentTree()) ix = 1;
      colName = mappingElement.columnNames()[ix]; // the position column is the first or second for dependent elements
      selColumn << colName<<"_"<<i;
      conditionData.extend<int>(selColumn.str());
      conditionData[selColumn.str()].data<int>() = selection.data()[iItem->first].data<int>();
    } else {
      const pool::ObjectRelationalMappingElement::const_iterator iElem = mappingElement.find("value_type");
      if ( iElem == mappingElement.end() ) {
        throw pool::RelationalStorageService::RelationalStorageServiceException( "Item for element \"value_type\" not found in the mapping element",
                                                                                 "ObjectReader::setQueryCondition" );
      }
      const pool::ObjectRelationalMappingElement& valueTypeElement = iElem->second;
      if( valueTypeElement.elementType()==pool::ObjectRelationalMappingElement::Primitive ){
        if(varName!="value_type"){
          throw pool::RelationalStorageService::RelationalStorageServiceException( "Item for element \"" + varName + "\" not found in the mapping element",
                                                                                   "ObjectReader::setQueryCondition" );
        }
        colName = valueTypeElement.columnNames()[0];
      } else if( valueTypeElement.elementType()==pool::ObjectRelationalMappingElement::Object ){
        pool::ObjectRelationalMappingElement::const_iterator iInnerElem = valueTypeElement.find(varName);
        if ( iInnerElem == valueTypeElement.end() ) {
          throw pool::RelationalStorageService::RelationalStorageServiceException( "Item for element \"" + varName + "\" not found in the mapping element",
                                                                                   "ObjectReader::setQueryCondition" );
        }
        colName = iInnerElem->second.columnNames()[0];
      } else {
        throw pool::RelationalStorageService::RelationalStorageServiceException( "Queries cannot be executed on types mapped on "+
                                                                                 pool::ObjectRelationalMappingElement::elementTypeAsString(valueTypeElement.elementType()),
                                                                                 "ObjectReader::setQueryCondition" );
      }
      selColumn << colName<<"_"<<i;
      conditionData.extend(selColumn.str(),selection.data()[iItem->first].specification().type());
      conditionData[selColumn.str()].setValueFromAddress(selection.data()[iItem->first].addressOfData());
    }
    cond << colName << " " << iItem->second << " :"<<selColumn.str();
    i++;
    selColumn.str("");
  }
  return cond.str();
}

      

