#include "RelationalStorageServiceException.h"
#include "RelationalStorageServiceNames.h"
#include "RelationalPoolDatabase.h"
#include "PoolDatabaseManager.h"
#include "PoolBaseContainerManager.h"
#include "StorageSvc/DbInstanceCount.h"
#include "ObjectRelationalAccess/ObjectRelationalMappingPersistency.h"
#include "ObjectRelationalAccess/ObjectRelationalMappingSchema.h"
#include "StorageSvc/DbDomain.h"
#include "StorageSvc/DbDatabase.h"
#include "StorageSvc/DbTransaction.h"
#include "POOLCore/Guid.h"
#include "POOLCore/Context.h"
#include "RelationalAccess/ConnectionService.h"
#include "RelationalAccess/ISessionProxy.h"
#include "RelationalAccess/ITransaction.h"
#include "RelationalAccess/IWebCacheControl.h"
#include "CoralBase/Exception.h"
#include "CoralBase/MessageStream.h"
#include <memory>

pool::RelationalStorageService::RelationalSession::RelationalSession(coral::ISessionProxy& session, bool owned):m_session(&session),
                                                                                                                m_owned(owned),
                                                                                                                m_myTransaction(false){
}

pool::RelationalStorageService::RelationalSession::~RelationalSession(){
  if(m_owned) delete m_session;
}

bool pool::RelationalStorageService::RelationalSession::startTransaction( bool readOnly ){
  bool started = false;
  coral::ITransaction& currentTransaction = m_session->transaction();
  if( !currentTransaction.isActive() ){
    currentTransaction.start( readOnly );
    m_myTransaction = true;
    started = true;
  }
  return started;
}

bool pool::RelationalStorageService::RelationalSession::commitTransaction(){
  coral::ITransaction& currentTransaction = m_session->transaction();
  if( currentTransaction.isActive() ){
    if(m_myTransaction){
      currentTransaction.commit();
      m_myTransaction = false;
      return true;
    }
  }
  return false;
}

bool pool::RelationalStorageService::RelationalSession::rollBackTransaction(){
  coral::ITransaction& currentTransaction = m_session->transaction();
  if( currentTransaction.isActive() ){
    if(m_myTransaction){
      if(!currentTransaction.isReadOnly()) currentTransaction.rollback();
      else currentTransaction.commit();
      m_myTransaction = false;
      return true;
    }
  }
  return false;
}

bool pool::RelationalStorageService::RelationalSession::transactionActive(){
  coral::ITransaction& currentTransaction = m_session->transaction();
  return currentTransaction.isActive();
}

coral::ISessionProxy* pool::RelationalStorageService::RelationalSession::get(){
  return m_session;
}

pool::RelationalStorageService::RelationalPoolDatabase::RelationalPoolDatabase( pool::IOODatabase* idb,
                                                                                pool::Context* context ):
  pool::DbDatabaseImp( idb ),
  m_formatVersion( "1.0" ),
  m_context( context ),
  m_ownedContext(false),
  m_session( 0 ),
  m_readOnlyMode( false ),
  m_databaseManager( 0 ),
  m_pfnName( "" )
{
  pool::DbInstanceCount::increment( this );
  if(!m_context) {
    m_context = new pool::Context;
    m_ownedContext = true;
  }
}


pool::RelationalStorageService::RelationalPoolDatabase::~RelationalPoolDatabase()
{
  if ( m_databaseManager ) delete m_databaseManager;
  if ( m_session ) {
    delete m_session;
  }
  pool::DbInstanceCount::decrement( this );
  if(m_ownedContext && m_context) delete m_context;
}

pool::Context*
pool::RelationalStorageService::RelationalPoolDatabase::context(){
  return m_context;
}

pool::RelationalStorageService::PoolDatabaseManager&
pool::RelationalStorageService::RelationalPoolDatabase::databaseManager()
{
  return *m_databaseManager;
}

pool::RelationalStorageService::RelationalSession*
pool::RelationalStorageService::RelationalPoolDatabase::getSession( const std::string& connectionString,
                                                                    coral::AccessMode accessMode,
                                                                    pool::Context* context ){
  // Retrieve an IConnectionService
  coral::IConnectionService* connectionService = context->connectionService();
  if(!connectionService){
    context->setConnectionService(new coral::ConnectionService());
    connectionService = context->connectionService();
  }
  configureConnectionService(*connectionService, connectionString);
  coral::ISessionProxy* session = 0;
  bool owned = false;
  if( context->isSessionSharingEnabled() ){
    session = &context->sharedSession( connectionString, accessMode );
  } else {
    session = connectionService->connect( connectionString, accessMode );
    owned = true;
  }
  if ( ! session ) {
    throw pool::RelationalStorageService::RelationalStorageServiceException("Could not connect to \""+connectionString+"\"",
                                                                            "RelationalPoolDatabaseUtilities::getSession");
  }
  return new RelationalSession( *session, owned );
}

bool
pool::RelationalStorageService::RelationalPoolDatabase::exists( const std::string& name,
                                                                pool::Context* context )
{
  coral::MessageStream log( pool::RelationalStorageService::serviceName() );
  if(!context){
    return false;
  }
  //
  bool existsTopLevelTable = false;
  RelationalSession* tempSession = 0;
  try{
    tempSession = getSession( name, coral::ReadOnly, context );
    if ( ! tempSession ) return false;
    tempSession->startTransaction( true );
    pool::RelationalStorageService::PoolDatabaseManager manager( tempSession->get(), context->properties() );
    existsTopLevelTable = manager.existsDatabase();
    tempSession->commitTransaction();
  }
  catch (const coral::Exception& e)
  {
    log << coral::Error << "Could not connect to \"" << name << "\"" << coral::MessageStream::endmsg;
    log << coral::Error << e.what()<< coral::MessageStream::endmsg;
    if(tempSession) delete tempSession;
    return false;
  }

  delete tempSession;
  return existsTopLevelTable;
}


coral::ISessionProxy&
pool::RelationalStorageService::RelationalPoolDatabase::session()
{
  return *m_session->get();
}


pool::DbStatus
pool::RelationalStorageService::RelationalPoolDatabase::onOpen( const pool::DbDatabase& dbH,
                                                                pool::DbAccessMode mode )
{
  coral::MessageStream log( pool::RelationalStorageService::serviceName() );
  std::string par_val = m_formatVersion;
  if ( !dbH.param( "FORMAT_VSN", par_val ).isSuccess() ) {
    if ( mode&pool::CREATE || mode&pool::UPDATE ) {
      dbH.addParam( "FORMAT_VSN", m_formatVersion );
    }
    else {
      log << coral::Warning << "No data format parameter present" << coral::MessageStream::endmsg;
    }
  }
  else {
    m_formatVersion = par_val;
  }

  try{
    m_session->commitTransaction();
  }
  catch (const coral::Exception& e)
  {
    if(m_session) delete m_session;
    m_session = 0;
    log << coral::Error << "Cannot commit transaction - RelationalPoolDatabase::onOpen"<<coral::MessageStream::endmsg;
    log << coral::Error << e.what() << coral::MessageStream::endmsg;
    return pool::Error;
  }
  
  return pool::Success;
}

pool::DbStatus
pool::RelationalStorageService::RelationalPoolDatabase::open( const pool::DbDomain&, 
                                                              const std::string& name, 
                                                              pool::DbAccessMode mode )
{
  coral::MessageStream log( pool::RelationalStorageService::serviceName() );
  if ( m_databaseManager ) delete m_databaseManager;
  if ( m_session ) {
    delete m_session;
    m_session = 0;
  }

  // Connect to the database
  if(!m_context){
    return pool::Error;
  }
  try {
    m_session = getSession( name, ( mode == pool::READ ) ? coral::ReadOnly : coral::Update , m_context );
  }
  catch (const coral::Exception& e) {
    log << coral::Error << "Cannot start the user session" << coral::MessageStream::endmsg;
    log << coral::Error << e.what() << coral::MessageStream::endmsg;
    if(m_session) delete m_session;
    m_session = 0;
    return pool::Error;
  }

  m_session->startTransaction( true );
  m_databaseManager = new pool::RelationalStorageService::PoolDatabaseManager( m_session->get(), m_context->properties() );
  m_session->commitTransaction();
  
  // Check for the existence of the pool database
  bool existsDatabase = pool::RelationalStorageService::RelationalPoolDatabase::exists( name, m_context );

  try
  {
      // Take the relevant actions depending on the opening mode and the existence of the database
    if ( existsDatabase ) {
      if ( mode == pool::CREATE ) {
        delete m_databaseManager;
        m_databaseManager = 0;
        if(m_session) delete m_session;
        m_session = 0;
        log << coral::Error << "Cannot create a database in \"" << name << "\". It already exists !!!" << coral::MessageStream::endmsg;
        return pool::Error;
      }
      else if ( mode == pool::RECREATE ) {
        m_session->startTransaction();
        m_databaseManager->dropDatabase();
        m_databaseManager->createDatabase();        
        m_session->commitTransaction();
      }
    }
    else {
      if ( mode == pool::READ ) {
        delete m_databaseManager;
        m_databaseManager = 0;
        if(m_session)delete m_session;
        m_session = 0;
        log << coral::Error << "Cannot open the non-existing database in \"" << name << "\" for reading" << coral::MessageStream::endmsg;
        return pool::Error;
      }
      m_session->startTransaction();
      m_databaseManager->createDatabase();
      m_session->commitTransaction();
    }
    // Set if necessary the read-only flag
    if ( mode == pool::READ ) m_readOnlyMode = true;
    else m_readOnlyMode = false;

    m_session->startTransaction( m_readOnlyMode );
    // I need to start the transaction here because the StorageSvc opens the special containers.
    // The transaction will be committed at the onOpen method.
  }
  catch (const coral::Exception& e)
  {
    delete m_databaseManager;
    m_databaseManager = 0;
    if(m_session) delete m_session;
    m_session = 0;
    log << coral::Error << "Cannot open the database in \"" << name << "\"." << coral::MessageStream::endmsg;
    log << coral::Error << e.what() << coral::MessageStream::endmsg;
    return pool::Error;
    
  }

  m_pfnName = name;
  return pool::Success;
}


pool::DbStatus
pool::RelationalStorageService::RelationalPoolDatabase::reopen( pool::DbAccessMode mode )
{
  coral::MessageStream log( pool::RelationalStorageService::serviceName() );
  if ( mode == pool::READ ) m_readOnlyMode = true;
  else m_readOnlyMode = false;

  try
  {     
    if ( m_session ) {
      delete m_session;
      m_session = 0;
    }

    // Connect to the database
    if(!m_context){
      return pool::Error;
    }
    try {
      m_session = getSession( m_pfnName, m_readOnlyMode ? coral::ReadOnly : coral::Update, m_context );
    }
    catch (const coral::Exception& e) {
      log << coral::Error << "Cannot start the user session" << coral::MessageStream::endmsg;
      log << coral::Error << e.what() << coral::MessageStream::endmsg;
      if(m_session) delete m_session;
      m_session = 0;
      return pool::Error;
    }

    m_session->startTransaction( m_readOnlyMode );
    m_databaseManager->reopen( m_session->get() );

  }
  catch(const coral::Exception& e)
  {
    if(m_session) delete m_session;
    m_session = 0;
    log << coral::Error << "Cannot reopen the database in \"" << m_pfnName << "\"." << coral::MessageStream::endmsg;
    log << coral::Error << e.what() << coral::MessageStream::endmsg;
    return pool::Error;    
  }
  
  return pool::Success;
}


pool::DbStatus
pool::RelationalStorageService::RelationalPoolDatabase::close( pool::DbAccessMode /* mode */ )
{
  try 
  {
    m_session->rollBackTransaction();
  }
  catch(const coral::Exception&)
  {
  }
  
  if ( m_databaseManager ) delete m_databaseManager;
  m_databaseManager = 0;
  if ( m_session ) delete m_session;
  m_session = 0;
  return pool::Success;
}


pool::DbStatus
pool::RelationalStorageService::RelationalPoolDatabase::transAct( pool::DbTransaction& refTransaction )
{
  coral::MessageStream log( pool::RelationalStorageService::serviceName() );
  try 
  {
    pool::Transaction::Action action = refTransaction.state();
    bool activeTransaction = m_session->transactionActive();
    switch ( action ) {
    case pool::Transaction::TRANSACT_START:
    case pool::Transaction::TRANSACT_ACTIVE:
      m_session->startTransaction( m_readOnlyMode );
      break;
    case pool::Transaction::TRANSACT_COMMIT: // This is the PersistencySvc commitAndHold()
      if ( ! activeTransaction ) {
        log << coral::Error << "There is no active transaction to commit." << coral::MessageStream::endmsg;
        return pool::Error;
      }
      if ( m_databaseManager ) m_databaseManager->signalEndOfTransaction();
      m_session->commitTransaction();
      m_session->startTransaction( m_readOnlyMode );
      break;
    case pool::Transaction::TRANSACT_FLUSH:
      if ( ! activeTransaction ) {
        log << coral::Error << "There is no active transaction to commit." << coral::MessageStream::endmsg;
        return pool::Error;
      }
      if ( m_databaseManager ) m_databaseManager->signalEndOfTransaction();
      m_session->commitTransaction();
      break;
    case pool::Transaction::TRANSACT_ROLLBACK:
    case pool::Transaction::TRANSACT_ENDED:
      if ( activeTransaction ) {
        if ( m_databaseManager ) m_databaseManager->signalEndOfTransaction();
        m_session->rollBackTransaction();
      }
      break;
    };
  }
  catch(const coral::Exception& e)
  {
    if(m_session) delete m_session;
    m_session = 0;
    log << coral::Error << "Cannot terminate the transaction. "<< coral::MessageStream::endmsg;
    log << coral::Error << e.what() << coral::MessageStream::endmsg;
    return pool::Error;
  }
  
  return pool::Success;
}

void
pool::RelationalStorageService::RelationalPoolDatabase::dropContainer( pool::Context* context,
                                                                       const std::string& databaseName,
                                                                       const std::string& containerName,
                                                                       bool removeDataTables ){

  // Connect to the database
  if(!context){
      throw pool::RelationalStorageService::RelationalStorageServiceException("Context not provided.",
                                                                              "RelationalPoolDatabaseUtilities::dropContainer");
  }
  RelationalSession* session = 0;
  try{
    session = getSession( databaseName, coral::Update, context );
    session->startTransaction();

    pool::RelationalStorageService::PoolDatabaseManager manager( session->get(), context->properties() );
    if(!manager.existsDatabase()){
      session->rollBackTransaction();
      delete session;
      throw pool::RelationalStorageService::RelationalStorageServiceException("No POOL Database found in "+databaseName+".",
                                                                              "RelationalPoolDatabaseUtilities::dropContainer");
    }

    // when requested, first remove all the materialized tables
    if(removeDataTables){
      pool::ObjectRelationalMappingPersistency mappingPersistency(session->get()->nominalSchema());
      std::vector<std::string> mappingsForContainer = mappingPersistency.mappingsForContainer(containerName);

      for(std::vector<std::string>::const_iterator iM=mappingsForContainer.begin();
          iM!=mappingsForContainer.end();++iM){
        mappingPersistency.removeMaterializedTables(*iM);
        mappingPersistency.removeMapping(*iM);
      }
    }

    // delete the container from the pool database 
    manager.dropContainer( containerName );
    
    session->commitTransaction();
    delete session;
    session = 0;
  }
  catch (const coral::Exception& /* e */)
  {
    if(session) delete session;
    throw pool::RelationalStorageService::RelationalStorageServiceException("Could not connect to \""+databaseName+"\"",
                                                                            "RelationalPoolDatabaseUtilities::dropContainer");
  }
  
  if(session) delete session;
}

void
pool::RelationalStorageService::RelationalPoolDatabase::configureConnectionService(coral::IConnectionService& service, const std::string& connectionString){
  coral::IWebCacheControl& cacheControl = service.webCacheControl();
  
  // set the mapping tables
  std::set<std::string> mappingTables = pool::ObjectRelationalMappingSchema::listTables();
  for(std::set<std::string>::const_iterator iT=mappingTables.begin();iT!=mappingTables.end();++iT){
    cacheControl.refreshTable(connectionString,*iT);
  }
  // set the pool db tables
  cacheControl.refreshTable(connectionString,RelationalStorageServiceNames::containerHeaderTableName());
  cacheControl.refreshTable(connectionString,RelationalStorageServiceNames::sequencesTableName());
  // set the pool private containers tables
  cacheControl.refreshTable(connectionString,RelationalStorageServiceNames::linksContainerName());
  cacheControl.refreshTable(connectionString,RelationalStorageServiceNames::shapesContainerName());
}

      
