#include <string>
#include <vector>
#include <algorithm>
#include <map>
#include "ArgQual.h"
#include "CmdLineArgs2.h"

/**********************************************************

CmdLineArgs2 is an extension of a map intended to map 
   qualifiers (e.g. -help) of a CLI to the index of 
   the argv[]. It's state is determined by a list of 
   ArgQual's/QualList. It then uses evalArgs to apply 
   those qualifiers to an argv[].

 - Qualifier -help is added in constructor
 - methods described in comments in definition

Example of Usage:

      Args2Container argsVec(thisProgram);

      QualList markers;
      markers.insert( make_pair("-somequal", ArgQual()) );
      markers["-somequal"].desc << "this is a qualifier with default=this";

      CmdLineArgs2 cmdLineArgs;
      cmdLineArgs.setArgQuals(markers);
      cmdLineArgs.evalArgs(argc,argv);
      argsVec.push_back(&cmdLineArgs);  // Add it to the list

      ...

**********************************************************/
using namespace std;

CmdLineArgs2::CmdLineArgs2( std::string name ) 
  : m_argsfine(false),m_id(name)
{	
   this->initTypes();
   // it should always have help
   m_quals.insert(std::make_pair("-help",ArgQual(0)));
}


CmdLineArgs2::CmdLineArgs2( QualList& quals, std::string name ) 
  : m_argsfine(false),m_quals(quals),m_id(name)
{	
   this->initTypes();
   // it should always have help
   m_quals.insert(std::make_pair("-help",ArgQual(0)));
}

void
CmdLineArgs2::initTypes()
{
   // Is there a way to do this with static something?
   m_allowedTypes.push_back("RelationalCollection");
   m_allowedTypes.push_back("RootCollection");
   m_allowedTypes.push_back("PHYSICAL_NAME");
   m_allowedTypes.push_back("LOGICAL_NAME");
   m_allowedTypes.push_back("GUID");
}


bool 
CmdLineArgs2::hasQual( const std::string arg ) const
{
  return ( this->find( arg ) != this->end() );
}

bool
CmdLineArgs2::evalArgs(std::vector<std::string>& argv) 
{
   m_argsfine = true;
   int argc=argv.size();
   // Markers are info stored in the base class map
   // Look for markers (anything beginning with '-') in argv and record their position
   std::vector < std::pair< int,std::string> > markPos;
   std::vector <std::string> marks;
   for (int i=1; i<argc; ++i) {
      if (argv[i][0]=='-') {
         markPos.push_back(std::make_pair(i,argv[i]));
         marks.push_back(argv[i]);
      }
   }

   // sort the marks by their position
   std::sort(markPos.begin(),markPos.end());

   // Check for required args
   for (QualList::iterator quit = m_quals.begin(); quit != m_quals.end(); ++quit) {
      if (
          quit->second.required && 
          std::find(marks.begin(),marks.end(),quit->first) == marks.end() 
         ) {
         std::cerr << "Must specify value for " << quit->first << std::endl;
         m_argsfine = false;
      }
   }

   // build a map of the begin,end positions for all marks
   std::vector < std::pair <int,std::string> >::iterator itmarks=markPos.begin();
   while (itmarks != markPos.end()) {
      std::string markerName=itmarks->second;
      int first,last;
      first = itmarks->first+1;
      ++itmarks;
      if (itmarks==markPos.end()) last=argc;
      else last=itmarks->first;
      //std::cout << markerName << " " << first << " " << last << std::endl;
      // reject ones that are not in qual list
      if ( m_quals.find(markerName) != m_quals.end() ) 
         this->insert( std::make_pair( markerName,std::make_pair(first,last) ) );
      //else std::cout << "CmdLineArgs: ERROR Unknown qualifier " << markerName << std::endl;
   }

   // Now check that they obey the rules and set flag accordingly
   std::map < std::string, std::pair <int,int> >::iterator mit = this->begin();
   while ( mit != this->end() ) {
      std::string qual = mit->first;
      int npars   = m_quals[qual].argsize;
      int diff = (*this)[qual].second - (*this)[qual].first;
      if (m_quals[qual].multiarg) {
         if ( abs(diff) < npars || diff%npars != 0) {
            std::cout << " ERROR Wrong number of parameters" << std::endl;
            std::cout << qual << " needs " << npars << " parameters"
                 << ", Found " << diff << std::endl;
            if ( this->find("-help") != this->end() ) {
               std::cout << "MORE INFO" << std::endl;
               m_quals[qual].print(std::cout); std::cout << std::endl;
            }
            m_argsfine = false;
         }
      }
      else {
         if (abs(diff)!=npars) {
            std::cout << " ERROR Wrong number of parameters" << std::endl;
            std::cout << qual << " needs " << npars << " parameters"
                 << ", Found " << diff << std::endl;
            if ( this->find("-help") != this->end() ) {
               std::cout << "MORE INFO" << std::endl;
               m_quals[qual].print(std::cout); std::cout << std::endl;
            }
            m_argsfine = false;
         }
      }
      ++mit;
   }

   return m_argsfine;
}

unsigned int
CmdLineArgs2::nArgs(std::string name)
{
   if (this->valid()) {
      CmdLineArgs2::iterator it = this->find(name);
      if (it != this->end()) 
         return abs(it->second.second - it->second.first)/m_quals[name].argsize;
      else return 0;
   }
   else {
      std::cerr << "nArgs called for invalid CmdLineArgs2 class" << std::endl;
      return 0;
   }
}

