/**
 * @file CollAppend.cpp
 * @brief Utility to append POOL collections to another POOL collection
 * @author K. Karr <Kristo.Karr@cern.ch>
 * @author Marcin.Nowak@cern.ch
 * $Id: CollAppend.cpp,v 1.61 2009/11/05 13:12:52 mnowak Exp $ 
 */

#include "CollectionBase/ICollection.h"
#include "CollectionBase/ICollectionFragment.h"
#include "CollectionBase/CollectionService.h"
#include "CollectionBase/CollectionDescription.h"
#include "CollectionBase/CollectionRowBuffer.h"
#include "CollectionBase/TokenList.h"
#include "CollectionBase/ICollectionColumn.h"
#include "CollectionBase/ICollectionQuery.h"
#include "CollectionBase/ICollectionCursor.h"
#include "CollectionBase/ICollectionDataEditor.h"
#include "CollectionBase/ICollectionMetadata.h"
#include "CollectionBase/CollectionBaseNames.h"

#include "CoralBase/MessageStream.h"
#include "POOLCore/Token.h"
#include "POOLCore/Exception.h"

#include "CoralBase/Attribute.h"

#include "Args2Container.h"
#include "DstInfo.h"
#include "SrcInfo.h"
#include "QueryInfo.h"
#include "CatalogInfo.h"
#include "MetaInfo.h"
#include "Progress.h"

#include <ctime>
#include <sstream>
#include <iostream>
#include <string>
#include <map>
#include <set>

#include <memory> 
#ifdef __GXX_EXPERIMENTAL_CXX0X__
//#   define AUTO_PTR  std::unique_ptr
#   define AUTO_PTR  auto_ptr
#else
#   define AUTO_PTR  auto_ptr
#endif

#define corENDL coral::MessageStream::endmsg

using namespace std;
using namespace pool;


int main(int argc, const char *argv[])
{
   // flag telling if there was any commit done
   bool	committed( false );
   
   // Primary try loop
   try
   {
      string thisProgram("CollAppend");
      // Copymode: affects how metadata is being copied
      enum CopyMode { copy, extract, merge } copyMode;
      copyMode = merge;
      string provName("SourceFID_ref");
      string TAALS_CLID("6F6A12A0-FEEF-484B-9691-94B82B90CDBA");
   
      bool noAttrib = false;
      bool noMetadata = false;
      // output modifiers
      long long numEvents = -1;
      //unsigned int numEventsPerCommit = 10000;
      unsigned int numEventsPerCommit = static_cast<unsigned int>(-1);
      int numEventsPerPrint = -1;
      int numRowsCached = 0;
      bool ignoreSchemaDiff = false;
      bool extendProv = false;
      
      std::vector<std::string> inputQuery;
      map<string,string> srcMetadata;
      time_t begin, end;
      // dst collection info
      std::vector<bool> dstCollExistVec;
     
      coral::MessageStream log( thisProgram );
      pool::CollectionService   collectionService;

      // Convert argv to vector of strings
      vector<string> argv_v;
      for( int i=0; i<argc; ++ i )
	 argv_v.push_back( argv[i] );
     
      // vector of CmdLineArg objects
     Args2Container argsVec( thisProgram, true, &log );
     argsVec.desc << "Takes one or more collections (of varying formats) and "
		  << "merges them and writes the results to one or more output "
		  << "collections (of varying formats). " << endl
		  << "Infile Metadata handling depends on -copy, -extract and -merge options." << endl;
 
      // list of CollAppend *specific* cli keys and their argument properties
      QualList markers;
      markers.insert( make_pair("-noattrib",     ArgQual(0)) );
      markers.insert( make_pair("-nevents",      ArgQual()) );
      markers.insert( make_pair("-nevtpercommit",ArgQual()) );
      markers.insert( make_pair("-nevtperprint", ArgQual()) );
      markers.insert( make_pair("-nevtcached",   ArgQual()) );
      // markers.insert( make_pair("-ignoreschemadiff",ArgQual(0)) );
      markers.insert( make_pair("-copy",        ArgQual(0)) );
      markers.insert( make_pair("-extract",     ArgQual(0)) );
      markers.insert( make_pair("-merge",       ArgQual(0)) );
      markers.insert( make_pair("-nometadata",  ArgQual(0)) );
      markers.insert( make_pair("-extendprov",  ArgQual()) );

      markers["-noattrib"].desc << "(no attributes copied, only event tokens) ";
      markers["-nevents"].desc << "<Max. number of rows to copy>";
      markers["-nevtpercommit"].desc << "[Max. number of events to process between "
		                     << "collection commits (DEFAULT = -1)]";
      markers["-nevtperprint"].desc << "Number of events to process before "
		                    << "each print of status to the screen.";
      markers["-nevtcached"].desc << "size of the insert buffer for bulk operations "
		                  << "DEFAULT = 0 (no bulk operations)";
      // markers["-ignoreschemadiff"].desc << "attempts to copy collections even if schema seems incompatible "
      //                                  << "(use at your own risk)";
      markers["-copy"].desc << "Make exact copy of 1 source collection ";
      markers["-extract"].desc << "Combine N source collections, prepend existing metadata keys with source collection name. "
                               << "Add new metadata without prefix, include metadata for: "
                               << "InputCollections, OutputCollections, Query, ExtractNode, TimeBegin, TimeEnd. "
                               << "(useful for creating standalone destination collections)";
      markers["-merge"].desc << "Combine N source collections, prepend existing metadata keys with source collection name. "
                             << "Add user supplied metadata with source collection name prefix. " 
                             << "(useful for appending to exiting destination collections). This is the DEFAULT. ";
      markers["-nometadata"].desc << "Ignore metadata. (will still accept new metadata from the command line -metadata option)";
      markers["-extendprov"].desc << "Add GUIDs of the input collections to the new Token attribute <arg>. "
				  << "If <arg> is \"default\" the default name will be used: " << provName ;

      
      CmdLineArgs2 cmdLineArgs;
      cmdLineArgs.setArgQuals(markers);
      argsVec.push_back(&cmdLineArgs);  // Add it to the list

      // Classes with shared cli keys and their argument properties
      // Add them to the list
      CatalogInfo catinfo; argsVec.push_back(&catinfo);
      QueryInfo queryinfo; argsVec.push_back(&queryinfo);
      DstInfo dstinfo; argsVec.push_back(&dstinfo);
      SrcInfo srcinfo; argsVec.push_back(&srcinfo);
      MetaInfo metainfo; argsVec.push_back(&metainfo);
      Progress progress; argsVec.push_back(&progress);

      // Print out help if requested
      if( argc < 3 ) {
         if (argc==2 && argv_v[1]=="-help") argsVec.printHelp(true);
         else  argsVec.printHelp(false);
         return 1;
      }

      // Check that all cmd line args are valid
      argsVec.evalArgs(argv_v);
      if (!argsVec.checkValid()) return 1;


      // Fill appropriate vectors based on CollAppend *specific* cmdLineArgs
      map < string, pair <int,int> >::iterator mit=cmdLineArgs.begin();
      // bool had_copymode = false;
      while( mit != cmdLineArgs.end() ) {
         string qual = mit->first;
         int ifirst = mit->second.first;
         // cout << "**********  processing option: " << qual << " " << argv_v[ifirst] << endl;
         //int ilast  = mit->second.second;
         //cout << "Parsing " << qual << " " << ifirst << " " << ilast << endl;
         if (qual == "-noattrib") {
	    noAttrib = true;
         }
         else if (qual == "-nevents") {
            istringstream  str( argv_v[ifirst] );
	    str >> numEvents;
         }
         else if (qual == "-nometadata") {
	    noMetadata = true;
         }
         else if( qual == "-copy" ) {
	    copyMode = copy;
            if( srcinfo.nSrc() > 1 ) {
               log << coral::Error << "Copy mode: 'copy' supports only one source collection"
                   << coral::MessageStream::endmsg;
               exit(2);
            }
         }
         else if (qual == "-extract") {
	    copyMode = extract;
         }
         else if (qual == "-merge") {
	    copyMode = merge;
         }

         // ---
         // EXPERT OPTIONS
         // For tuning purposes the number of events between commits can be specified
         // ---
         else if (qual == "-nevtpercommit") {
	    numEventsPerCommit = atoi( argv_v[ifirst].c_str() );
         }

         else if (qual == "-nevtperprint") {
            std::string numEventsPerPrintStr = argv_v[ifirst];
	    if( numEventsPerPrintStr == "all" ) {
	       numEventsPerPrint = -1;
	    }
	    else if( numEventsPerPrintStr.size() && numEventsPerPrintStr[0] != '-' )  {
	       numEventsPerPrint = atoi( argv_v[ifirst].c_str() );
	    }
         }
	 else if( qual == "-extendprov" ) {
            extendProv = true;
            if( argv_v[ifirst] != "default" )
	       provName = argv_v[ifirst];  // + "_ref";
         }
         else if (qual == "-nevtcached") {
	    numRowsCached = atoi( argv_v[ifirst].c_str() );
         }

         else if (qual == "-ignoreschemadiff") {
	    ignoreSchemaDiff = true;
         }
         else {
            log << coral::Error << "Unrecognized option: " << qual << endl
                << "Use -help for option list" << coral::MessageStream::endmsg;
            exit(-1); 
         }

         ++mit;
      } 
     
      // Timing start HERE
      time(&begin);
      catinfo.setCatalogs( &collectionService );
  

/// INTRO COMPLETE

/// BEGIN PROCESSING

      std::map<unsigned int, pool::ICollection*> dstCollMap;
      // ---
      // Now loop over destination collections
      //   - possibly moot since this is normally one destination
      // ---
      
      // initialize the exist vector
      for (unsigned int j=0; j<dstinfo.nDst(); ++j) dstCollExistVec.push_back(true);


      for( unsigned int i=0; i < dstinfo.nDst(); ++i ) {
	 log << "Testing collection '" << dstinfo.name(i)
	     << "' of type " << dstinfo.type(i) <<  coral::MessageStream::endmsg;
	 dstCollMap[i] = 0;
	 // testing for collection existence
	 try {
	    bool readOnly( true );
	    AUTO_PTR<pool::ICollection> dstCollection( collectionService.handle( dstinfo.name(i), dstinfo.type(i), dstinfo.connect(), readOnly ) );

	    AUTO_PTR<pool::ICollectionQuery> dquery( dstCollection->newQuery() );
	    pool::ICollectionCursor& cursor = dquery->execute();
	    if( !cursor.next() ) {
	       log << coral::Info
		   << "Destination collection `" + dstinfo.name(i) + "' is empty."
		   <<  coral::MessageStream::endmsg;
	    }
	 }
	 catch ( pool::Exception& ) {
            if( copyMode == merge ) log << coral::Warning
                                        << "Destination Collection `" << dstinfo.name(i)
                                        << "' does not exist; A new one will be created; "
                                        << coral::MessageStream::endmsg;
	    dstCollExistVec[i] = false;
	 }
         if( dstCollExistVec[i] && copyMode != merge ) {
            log << coral::Error << "Destination collection '" << dstinfo.name(i)
                << "' exists, and the copy mode was not 'merge'. "
                << "Remove the existing collection first"
                << coral::MessageStream::endmsg;
            exit(1);
         }
      }
      

      // src collection info
      std::vector<int> 	srcCountVec;
      unsigned int 	evtCounterTotal = 0;
      std::map<int, pool::ICollection*> srcCollMap;
      // user to mark provenance columns removed from the source
      std::set<string>	removedTokens;    
      bool 		provCollExists = false;

      // ---
      // Now create ICollection for source collection
      // *AND* check whether the schemas are equivalent
      // ---
      size_t k=0;
      log << coral::Info << "Number of input collections: " << srcinfo.nSrc()
          << coral::MessageStream::endmsg;      
      for( unsigned int sCollN=0; sCollN < srcinfo.nSrc(); sCollN++ )
      {
	 progress.print("Opening source collection", 100*sCollN/srcinfo.nSrc());
	 log << coral::Info
	     << "Opening source collection " << srcinfo.name(sCollN)
	     << " of type " << srcinfo.type(sCollN) <<  coral::MessageStream::endmsg;
	 bool readOnly( true );
	 pool::ICollection *srcCollection = collectionService.handle( srcinfo.name(sCollN), srcinfo.type(sCollN), srcinfo.connect(), readOnly );

         //Try to register it in the catalog
         if (catinfo.useCC()) {
            try {
	       progress.print("Registering source collection", 0);
	       log << coral::Debug
		   <<  "----  Registering source collection " << srcinfo.name(sCollN)
		   <<  coral::MessageStream::endmsg;
               collectionService.registerExisting( srcinfo.name(sCollN),
                                                   srcinfo.type(sCollN),
                                                   srcinfo.connect(),
						   srcinfo.name(sCollN));
            } catch(pool::Exception& e) {
               log << coral::Info << " Source collection registration failed: " << e.what()
		   << corENDL;
            }
         }

	 srcCollMap[sCollN] = srcCollection;
         srcCountVec.push_back(0);
	 log << "Opened source collection " << srcinfo.name(sCollN) << corENDL;

	 string srcGUID = "NOTFOUND" ;
	 if( ( extendProv || !noMetadata )
	     && srcCollection->metadata().existsKey(CollectionBaseNames::CollIDMdataKey()) )
	    srcGUID = srcCollection->metadata().getValueForKey(CollectionBaseNames::CollIDMdataKey());
	 log << coral::Debug << "Collection " <<srcinfo.name(sCollN) << " GUID= "
	     << srcGUID << corENDL;
	 
         // Get schema, define first as element 0
	 const pool::ICollectionDescription &firstDesc = srcCollMap[0]->description();
         // Get schema for current collection defined by index i
	 const pool::ICollectionDescription &currentSrcDesc = srcCollection->description();

	 log << "Creating query for the source collection" <<  coral::MessageStream::endmsg;
	 AUTO_PTR<ICollectionQuery>  srcQuery( srcCollection->newQuery() );
	 srcQuery->setRowCacheSize( 1000 );
	 srcQuery->setCondition( queryinfo.query() );
	 if( queryinfo.queryOptions().size() ) {
	    srcQuery->addToOutputList( queryinfo.queryOptions() );
	 } else {
            if( noAttrib )
               srcQuery->selectAllTokens();
            else
               srcQuery->selectAll();
	 }
	 log << "Executing query for the source collection" <<  coral::MessageStream::endmsg;
	 progress.print("Querying source collection", 0);
	 pool::ICollectionCursor& cursor = srcQuery->execute();	 
 
         if( sCollN == 0 ) {
            // Prepare destination collections

            // Create Description of the destination collection
            // use only selected columns if this option was specified
            CollectionDescription	destDescription( srcinfo.name(sCollN), "type" );
	    const TokenList &tokens =  cursor.currentRow().tokenList();
	    for( TokenList::const_iterator token = tokens.begin();
		 token != tokens.end();  ++token ) {
	       string name = token.tokenName();	
	       const ICollectionColumn& column = currentSrcDesc.column( name );
	       if( name == firstDesc.eventReferenceColumnName() ) {
		  if( name == "Token" ) {
		     /* MN: hack - "Token" copied from RootCollection*/
		     // remap back the RootCollectino default event ref name to the new default
		     name = pool::CollectionBaseNames::defaultEventReferenceColumnName();
		  } else {
		     // source had user defined name, copy it
		     destDescription.setEventReferenceColumnName( name );
		  }
	       }
	       // Do not copy PROV columns if adding a new PROV columns
	       if( column.annotation().find("PROV") == string::npos || !extendProv ) {
		  destDescription.insertTokenColumn( name, column.annotation(),
						     column.collectionFragmentName() );
	       } else {
		  log << coral::Debug << "Ignoring source collection provenance Token attribute: "
		      << name << corENDL;
		  removedTokens.insert( name );
	       }
	    }
	    // add the new provenance Token
	    if( extendProv ) try {
	       destDescription.insertTokenColumn( provName, "PROV",
						  destDescription.collectionFragment(0).name() );
	       provCollExists = true;
	    }catch( pool::Exception &e ) {
	       log << coral::Error << "Failed to add provenance Token attribute " << provName
		   << "; " << e.what() << corENDL;
	    }
     
	    const coral::AttributeList &attribs = cursor.currentRow().attributeList();
	    for( coral::AttributeList::const_iterator attrib = attribs.begin();
		 attrib != attribs.end();  ++attrib ) {
	       const string& name = attrib->specification().name();
	       const ICollectionColumn& column = currentSrcDesc.column( name );
	       destDescription.insertColumn( name, column.type(), column.annotation(),
					     column.collectionFragmentName(),
					     column.maxSize(), column.sizeIsFixed() );
	    }

            // open destination collections
            for( unsigned int i=0; i<dstinfo.nDst(); i++ ) {
	       progress.print("Opening destination collections", 100*i/dstinfo.nDst());
	       if( dstCollExistVec[i] ) {
                  if( !dstCollMap[i] ) {
                     bool readOnly(false);
                     dstCollMap[i] = collectionService.handle( dstinfo.name(i), 
							       dstinfo.type(i), 
							       dstinfo.connect(), 
							       readOnly );
                     if (catinfo.useCC()) {
                        // Try to register it in the catalog
                        try { 
			   log << coral::Debug
			       <<  "----  Registering destination collection " << dstinfo.name(i) 
			       <<  coral::MessageStream::endmsg;
                           collectionService.registerExisting( dstinfo.name(i),
                                                               dstinfo.type(i),
                                                               dstinfo.connect() );
                        } catch(pool::Exception& e) {
                           log << coral::Warning << "Registration of collection "
			       << dstinfo.name(i) << " failed with error: " << e.what() << corENDL;
                        }
                     }
                  } else {
                     // For each source collection get clean dest. coll. row buffers
                     dstCollMap[i]->dataEditor().clearRowBuffers();
                  }
                  // Check whether an opened dst collection has an appropriate description 
                  if( !destDescription.equals( dstCollMap[i]->description() ) ) {
                     cerr << "CollAppend: ERROR -" 
                          << " Destination collection '" << dstCollMap[i]->description().name()
                          << "' does not have all columns requested from source collection '"
                          << currentSrcDesc.name() << "'" << endl;
                     cerr << " ABORTING " << endl;
                     exit(1);
                  }
               }
            }
         
            // Create/open destination collections  
            for ( unsigned int i=0; i<dstinfo.nDst(); i++ ) {
               if( !dstCollExistVec[i] ) {
                  // create missing destination collections
		  progress.print("Creating destination collections", 100*i/dstinfo.nDst());
		  std::string dstconnect = dstinfo.connect();
                  pool::CollectionDescription newDestDesc( destDescription );
                  newDestDesc.setName( dstinfo.name(i) );
                  newDestDesc.setType( dstinfo.type(i) ); 
                  newDestDesc.setConnection( dstconnect );
                  log << coral::Info << "Creating destination collection "
                      << dstinfo.name(i)
                      << " of type " << dstinfo.type(i) <<  coral::MessageStream::endmsg;

                  try { 
		     bool overwrite( false );  // don't overwrite existing collection
		     dstCollMap[i] = collectionService.create( newDestDesc, overwrite );
                     // created the collection - now try to register it
                     if(catinfo.useCC()) {
                        bool reg_ok = false;
			log << coral::Debug
			    <<  "----  Registering destination collection " << dstinfo.name(i) 
			    <<  coral::MessageStream::endmsg;
			try {
			   bool overwrite( true );  // overwrite existing catalog entry
			   reg_ok = collectionService.registerExisting( dstCollMap[i], overwrite, dstinfo.name(i) ); 
			} catch(pool::Exception& e) {
			   log << coral::Error << e.what() << coral::MessageStream::endmsg;
			}
			if( !reg_ok ) {
			   log << coral::Warning << "Registration of destination collection "
			       << dstinfo.name(i) 
			       << " FAILED!" << coral::MessageStream::endmsg;
			}
		     }	
                  } catch( pool::Exception& ) {
                     log << coral::Info <<"Could not create destination collection "
                         << dstinfo.name(i)
                         << " - testing again if the collection exists already " << coral::MessageStream::endmsg;
                     if( collectionService.exists( dstinfo.name(i), dstinfo.type(i), dstconnect ) ) {
                        bool forUpdate(false);
                        log << coral::Info <<"Opening destination collection "
                            << dstinfo.name(i)
                            << " in update mode" << coral::MessageStream::endmsg;
                        dstCollMap[i] = collectionService.handle( dstinfo.name(i),
                                                                   dstinfo.type(i),
                                                                   dstconnect, forUpdate );
                     } else {  
                        throw;
                     }
                  }
                  dstCollExistVec[i] = true;
               }
               // set Row Cache size for ALL collections here
               if( numRowsCached )
                  dstCollMap[i]->dataEditor().setRowCacheSize( numRowsCached );
            }
         }

         // end of preparing destination collections
      
         
         // ---
         // Now to the meat of the program
         // - get an iterator to the input
         // - loop over it
         // - fill the output
         // ---
	 pool::Token provToken;
	 provToken.setDb( srcGUID );
	 provToken.setCont( "POOLCollection" );
	 provToken.setClassID( pool::Guid(TAALS_CLID) );
	 provToken.setTechnology( 0 );
	       
         log << "Iterating over the source collection" <<  coral::MessageStream::endmsg;
         long long evtCounter = 0;
         const std::string& srcRefName = srcCollection->description().eventReferenceColumnName();
	 size_t		srcSize = 1;  // the size of the query result set is not known - todo sometime
	 size_t		progressStep = (srcSize>99) ? srcSize/100 : 1;
	 
         while( cursor.next() && (numEvents < 0 || numEvents > evtCounter) ) {
            evtCounter++;
            evtCounterTotal++;
	    if( evtCounter % progressStep )
	       progress.print("Copying collection(s)", 100*evtCounter/srcSize);

            for( unsigned int i=0; i<dstinfo.nDst(); i++ )
            {
               // Get row buffer for adding data to collection.
	       pool::CollectionRowBuffer &rowBuffer = dstCollMap[i]->dataEditor().rowBuffer();

               // copy attributes
               rowBuffer.attributeList() = cursor.currentRow().attributeList();
               // copy tokens
               const std::string& dstRefName = dstCollMap[i]->description().eventReferenceColumnName();
               const pool::TokenList&  tokens = cursor.currentRow().tokenList();
               for( pool::TokenList::const_iterator ti = tokens.begin(); ti != tokens.end(); ++ti ) {
		  if( !extendProv || removedTokens.find( ti.tokenName() ) == removedTokens.end() ) {
		     if( ti.tokenName() == srcRefName ) {
			ti->setData( &rowBuffer.tokenList()[ dstRefName ] );
		     } else {
			ti->setData( &rowBuffer.tokenList()[ ti.tokenName() ] );
		     }
                  }
               }
	       if( provCollExists ) {
		  provToken.setData( &rowBuffer.tokenList()[provName] );
	       }
	       dstCollMap[i]->dataEditor().insertRow( rowBuffer );

	       if( !(evtCounterTotal % numEventsPerCommit) ) {
		  dstCollMap[i]->commit();
		  committed = true;
	       }

	       if ( (numEventsPerPrint >= 0) && !(evtCounterTotal % numEventsPerPrint) ) {
		  log << evtCounterTotal << " events appended" <<  endl;
	       }
	    }
	 }
         if( k < srcCountVec.size() ) srcCountVec[k]=evtCounter;
         k++;

         
//-------  PREPARE METADATA
	 progress.print("Metadata processing", 0);

	 // Stash all input metadata in a map
         string srcCollName = srcCollMap[sCollN]->description().name();
         ICollectionMetadata &srcMeta =  srcCollMap[sCollN]->metadata();
         if( srcMeta.existsKey("Name") ) {
            srcCollName = srcMeta.getValueForKey("Name");
            log << coral::Debug << "Input collection "
                << srcCollMap[sCollN]->description().name()
                << " has internal metadata Name=" << srcCollName
                << coral::MessageStream::endmsg;
         }

         bool meta_exists = false;
         if( !noMetadata && copyMode == merge ) {
            for( size_t i=0; i<dstinfo.nDst(); i++ ) {
               ICollectionMetadata &dstMeta =  dstCollMap[i]->metadata();
               for( ICollectionMetadata::const_iterator it = dstMeta.begin(); it != dstMeta.end(); ++it) {
                  if( it.key().compare(0, srcCollName.length()+1, srcCollName+"|" ) == 0 ) {
                     meta_exists = true;
                     break;
                  }
               }
               if( meta_exists ) {
                  log << coral::Warning << "*** Metadata from source collection "
                      << srcCollName << " already exists in destination collection "
                      << dstCollMap[i]->description().name()
                      << ". It will NOT be copied to any collection."
                      << coral::MessageStream::endmsg;
                  break;
               }
            }
         }

         if( !meta_exists ) {
            if( !noMetadata ) {
               for( ICollectionMetadata::const_iterator sit = srcMeta.begin(); sit != srcMeta.end(); ++sit) {
                  string key = sit.key();
                  if( copyMode != copy )
                     key =  srcCollName + "|" + key;
                  srcMetadata[ key ] = sit.value();
               }
               log << coral::Info << "Input collection " << srcCollName
                   << " has " <<  srcMetadata.size() << " metadata entries"
                   << coral::MessageStream::endmsg;

               // Check if input is previously derived collection
               if( queryinfo.hasQual("-query") && 
                   srcCollMap[sCollN]->metadata().existsKey("Query") ) {
                  log << coral::Warning << srcinfo.name(sCollN) << " is already derived" << endl
                      << "  ==> Found Query = " 
                      << srcCollMap[sCollN]->metadata().getValueForKey("Query") 
                      << coral::MessageStream::endmsg;
               }
            }
            // add new metadata from the commandline
            //  ignores -nometadata 
            for( size_t mk=0; mk<metainfo.nMeta(); mk++ ) {
               string key = metainfo.getKeyN(mk);
               if( copyMode == merge ) {
                  // prepend source collection name
                  key = srcCollName + "|" + key;
               }
               srcMetadata[ key ] = metainfo.getValueN(mk);
               log << coral::Debug << "** adding metadata " << key
                   <<"=" << metainfo.getValueN(mk) << coral::MessageStream::endmsg;
            }      
         } // ------  END METADATA
         
      }
      // Time end here
      time(&end); 
   

// ----------  STORE METADATA
      progress.print("Metadata writing", 0);
      // add extract information
      if( !noMetadata && copyMode == extract ) {
         string user; if (getenv("USER")!=NULL) user = getenv("USER");
         string host; if (getenv("HOST")!=NULL) host = getenv("HOST");
         string userhost = user + "@" + host;
         stringstream starttime,endtime;

         std::string inputCollections  = srcinfo.name(0);
         for( size_t i=1; i<srcinfo.nSrc(); i++ ) 
            inputCollections += ":"+srcinfo.name(i);
         std::string outputCollections = dstinfo.name(0);
         for( size_t i=1; i<dstinfo.nDst(); i++ ) 
            outputCollections += ":"+dstinfo.name(i);
         starttime << begin; endtime << end;

         srcMetadata[ "TimeBegin" ] = starttime.str();
         srcMetadata[ "TimeEnd" ] = endtime.str();
         srcMetadata[ "ExtractNode" ] = userhost;
         srcMetadata[ "InputCollections" ] = inputCollections;
         srcMetadata[ "OutputCollections" ] = outputCollections;
         srcMetadata[ "Query" ] = (queryinfo.query().length() ? queryinfo.query() : "1=1");
      }
      
      // write all metadata
      for( size_t i=0; i<dstinfo.nDst(); i++ ) {
         for( map<string,string>::const_iterator mit = srcMetadata.begin();
              mit != srcMetadata.end(); ++mit ) {
            if (!dstCollMap[i]->metadata().existsKey(mit->first)) 
               dstCollMap[i]->metadata().setValueForKey(mit->first,mit->second);
            else {
	       if( mit->first != CollectionBaseNames::CollIDMdataKey() )  // ignore GUID key
		  cout << "WARNING: DST name = " << dstinfo.name(i) 
		       << " already has key = " << mit->first << endl; 
            }
         }

         dstCollMap[i]->commit();
	 committed = true;
         dstCollMap[i]->close();
         delete dstCollMap[i];
      }

      
/// FINISHED PROCESSING
/// CLEANUP
      for( size_t i=0; i<srcinfo.nSrc(); i++ ) {
         srcCollMap[i]->close();
	 delete srcCollMap[i];
      }

      
      // List individual collections (with evt counts)
      cout << "CollAppend: Finished appending input collection(s): " << endl;
      for( unsigned int i=0; i<srcinfo.nSrc(); i++ ) {
	 cout << i+1 << ": `" << srcinfo.name(i) << ":" 
                              << srcinfo.type(i) << "' (" 
                           << srcCountVec[i]    << " events) "
                           << endl;
      }
      
      // List individual output collections
      cout << " to output collection(s): " << endl;
      for( unsigned int i=0; i<dstinfo.nDst(); i++ ) {
	 cout << i+1 << ": `" << dstinfo.name(i) << ":" 
	                      << dstinfo.type(i) << "' "
                           << endl;
      }
      
     // Note if there was a selection involved
      if (queryinfo.query().size() > 0) cout << "Qualified by " << queryinfo.query() << endl;
      // Give total for output
      cout << endl;
      cout << " ===> " << evtCounterTotal << " total events appended to each output" << endl;
      cout << " ===> " << "Processing took " << difftime(end,begin) << " seconds" << endl;
      if( committed ) cout << "At least one COMMIT has occurred" << endl;

      return 0;
   }
   catch( pool::Exception& poolException )
   {
      if( committed ) cout << "At least one COMMIT has occurred" << endl;
      std::cerr << "pool::Exception: " << poolException.what() << std::endl;;
      return 1;
   }
   catch( std::exception& exception )
   {
      if( committed ) cout << "At least one COMMIT has occurred" << endl;
      std::cerr << "std::exception: " << exception.what() << std::endl;;
      return 1;
   }
   catch( ... )
   {
      if( committed ) cout << "At least one COMMIT has occurred" << endl;
      std::cerr << "Unknown exception caught in main()." << std::endl;;
      return 1;
   }
}
