/**
 * @file CollListFileGUID.cpp
 * @brief Utility to list the file GUID's used by a POOL collection
 * @author K. Karr <Kristo.Karr@cern.ch>
 * @author Marcin.Nowak@cern.ch
 * $Id: CollListFileGUID.cpp,v 1.31 2009/10/30 11:09:50 mnowak Exp $
 */

#include "CollectionBase/ICollection.h"
#include "CollectionBase/TokenList.h"
#include "CollectionBase/ICollectionQuery.h"
#include "CollectionBase/ICollectionCursor.h"
#include "CollectionBase/CollectionRowBuffer.h"
#include "CollectionBase/CollectionService.h"
#include "CollectionBase/CollectionDescription.h"

#include "PersistencySvc/ISession.h"
#include "PersistencySvc/DatabaseConnectionPolicy.h"
#include "PersistencySvc/ITransaction.h"

#include "CoralBase/MessageStream.h"
#include "POOLCore/Token.h"
#include "POOLCore/Exception.h"

#include "Args2Container.h"
#include "SrcInfo.h"
#include "QueryInfo.h"
#include "CatalogInfo.h"

#include <iostream>
#include <iomanip>
using namespace std;


int main(int argc, const char *argv[])
{
   string thisProgram("CollListFileGUID");
   bool		mainToken( false );
   bool		doCounts( false );
     
   // Convert argv to vector of strings
   vector<string> argv_v;
   for( int i=0; i<argc; ++ i )
      argv_v.push_back( argv[i] );

   try
   { 
      coral::MessageStream log( thisProgram );
      pool::CollectionService   collectionService;

      // vector of CmdLineArg objects
      Args2Container argsVec(thisProgram, true, &log);
      argsVec.desc << thisProgram
                   << " is a tool for listing POOL file GUIDs referenced from the source event collection(s). "
                   << endl;

      QualList markers;
      markers.insert( make_pair("-maintoken", ArgQual(0)) );
      markers["-maintoken"].desc << "Print only GUIDs from the primary collection Token";
      markers.insert( make_pair("-counts",    ArgQual(0)) );
      markers["-counts"].desc << "Provide event counts for each GUID, slower";
	  
      CmdLineArgs2 cmdLineArgs;
      cmdLineArgs.setArgQuals(markers);
      argsVec.push_back(&cmdLineArgs); 

      CatalogInfo catinfo; argsVec.push_back(&catinfo);
      QueryInfo queryinfo; argsVec.push_back(&queryinfo);
      SrcInfo srcinfo;     argsVec.push_back(&srcinfo);

      // Print out help if requested
      if( argc < 3 ) {
         if (argc==2 && argv_v[1]=="-help") argsVec.printHelp(true);
         else  argsVec.printHelp(false);
         return 1;
      }

      // Check that all cmd line args are valid
      argsVec.evalArgs(argv_v);
      if (!argsVec.checkValid()) return 1;

      // Fill appropriate vectors based on *specific* cmdLineArgs
      map < string, pair <int,int> >::iterator mit=cmdLineArgs.begin();
      while( mit != cmdLineArgs.end() ) {
         string qual = mit->first;
	 if( qual == "-maintoken" ) {
	    mainToken = true;
         } else if( qual == "-counts" ) {
	    doCounts = true;
	 }
	 else {
	    //int ifirst = mit->second.first;
	    log << coral::Error << "Unrecognized option: " << qual << endl
		<< "Use -help for option list" << coral::MessageStream::endmsg;
	    exit(-1);
	 }
	 ++mit;
      }

      if( mainToken && queryinfo.queryOptions().size() ) {
	 log << coral::Error << "-maintoken and -queryopt flags are exclusive"
	     << coral::MessageStream::endmsg;
         exit(-2); 
      }

      catinfo.setCatalogs( &collectionService );

      std::map<int, pool::ICollection*> collMap;
      std::map< std::string, int > 	fileIdMap;
      std::map< std::string, std::string> guidtoNameMap;
      for( unsigned int i=0; i<srcinfo.nSrc(); i++ ) {

	 bool readOnly( true );
         pool::ICollection *collection = collectionService.handle( srcinfo.name(i), srcinfo.type(i), srcinfo.connect(), readOnly );
	 collMap[i] = collection;

	 if( mainToken ) {
	    log << coral::Info << "Collection: " << collection->description().name()
		<< ", Primary Token name is: " << collection->description().eventReferenceColumnName()
		<< coral::MessageStream::endmsg; 
	 }
	    
	 log << coral::Debug
             << ">> Creating query for the collection" << coral::MessageStream::endmsg;            
	 pool::ICollectionQuery *collQuery = collection->newQuery();
         collQuery->setCondition( queryinfo.query() );
         if( mainToken ) {
	    // get the main token only by default
	 } else if( queryinfo.queryOptions().size() ) {
	    collQuery->skipEventReference( true );
            collQuery->addToOutputList( queryinfo.queryOptions() );
         } else {
	    collQuery->selectAllTokens();
         }
	 log << coral::Debug
	     << ">> Executing the query " << coral::MessageStream::endmsg;
	 pool::ICollectionCursor& cursor = collQuery->execute();

	 while( cursor.next() ) {
	    const pool::TokenList &tokens = cursor.currentRow().tokenList();
	    for( pool::TokenList::const_iterator iter = tokens.begin();
		 iter != tokens.end(); ++iter ) {
	       std::string fileId = iter->dbID();
	       std::map< std::string, int >::iterator mapIter;
	       mapIter = fileIdMap.find( fileId );
	       if( mapIter == fileIdMap.end() ) {
		  fileIdMap[ fileId ] = 1;
		  guidtoNameMap[ fileId ] = iter.tokenName();
	       } else {
		  ++(mapIter->second);
	       }
	    }
	 }

	 std::map<std::string,int>::iterator it = fileIdMap.begin();
	 while( it!=fileIdMap.end() ) {
	    std::string tName = "UNKNOWN";
	    std::map<std::string,std::string>::iterator name_it = guidtoNameMap.find(it->first);
	    if( name_it != guidtoNameMap.end() )
	       tName = name_it->second;
	    std::cout << it->first << " " << setw(20) << setiosflags(ios::left) << tName;
	    if( doCounts )
	       cout << "  (" << it->second << ")";
	    cout << std::endl;
	    ++it;
	 }
      }
      for( unsigned int i=0; i<srcinfo.nSrc(); i++ ) {      
	 collMap[i]->close();
      }
      return 0;

   }
   catch( pool::Exception& poolException )
   {
      std::cerr << "pool::Exception: " << poolException.what() << std::endl;
      return 1;
   }
   catch( std::exception& exception )
   {
      std::cerr << "std::Exception: " << exception.what() << std::endl;
      return 1;
   }
   catch( ... )
   {
      std::cerr << "Unknown exception. " << std::endl;
      return 1;
   }
}
