/**
 * @file CollRemove.cpp
 * @brief Utility to delete a POOL collection
 * @author Marcin.Nowak@cern.ch
 */
#include "CollectionBase/ICollectionDescription.h"
#include "CollectionBase/CollectionService.h"
#include "CollectionBase/ICollection.h"

#include "FileCatalog/IFileCatalog.h"
#include "FileCatalog/IFCAction.h"

#include "CoralBase/MessageStream.h"
#include "POOLCore/Exception.h"

#include "Args2Container.h"
#include "SrcInfo.h"
#include "CatalogInfo.h"

#include <iostream>
using namespace std;

int main(int argc, const char *argv[])
{
   string thisProgram("CollRemove");

   // Convert argv to vector of strings
   vector<string> argv_v;
   for( int i=0; i<argc; ++ i )
      argv_v.push_back( argv[i] );

   try
   { 
      coral::MessageStream log( thisProgram );
      pool::CollectionService   collectionService;

      // vector of CmdLineArg objects
      Args2Container argsVec(thisProgram, true, &log);
      argsVec.desc << thisProgram
                   << " is a tool for deleting Relational POOL Collections. "
                   << endl;

      QualList markers;
      CmdLineArgs2 cmdLineArgs;
      cmdLineArgs.setArgQuals(markers);
      argsVec.push_back(&cmdLineArgs); 

      CatalogInfo catinfo; argsVec.push_back(&catinfo);
      SrcInfo srcinfo;     argsVec.push_back(&srcinfo);

      // Print out help if requested
      if( argc < 3 ) {
         if (argc==2 && argv_v[1]=="-help") argsVec.printHelp(true);
         else  argsVec.printHelp(false);
         return 1;
      }

      // Check that all cmd line args are valid
      argsVec.evalArgs(argv_v);
      if (!argsVec.checkValid()) return 1;

      // Fill appropriate vectors based on *specific* cmdLineArgs
      map < string, pair <int,int> >::iterator mit=cmdLineArgs.begin();
      while( mit != cmdLineArgs.end() ) {
         string qual = mit->first;
         //int ifirst = mit->second.first;
         log << coral::Error << "Unrecognized option: " << qual << endl
             << "Use -help for option list" << coral::MessageStream::endmsg;
         exit(-1); 
         ++mit;
      } 

      catinfo.setCatalogs( &collectionService );

      std::map<int, pool::ICollection*> collMap;
      std::map< std::string, int > pfnMap;
      for( unsigned int i=0; i<srcinfo.nSrc(); i++ ) {
         if( srcinfo.type(i)=="RootCollection" ) {
            log << coral::Error
                << "This tool only supports deleting Relational Collections"
                << coral::MessageStream::endmsg;
            return 2;
         }
         log << coral::Info
             << "Removing collection '" << srcinfo.name(i)
             << "', type: " <<  srcinfo.type(i)
             << coral::MessageStream::endmsg;

         bool deleteChildFragments = true;
         bool ignoreExternalDependencies = true;   // ?? MN
         bool result = collectionService.drop( srcinfo.name(i), srcinfo.type(i), srcinfo.connect(), deleteChildFragments, ignoreExternalDependencies);
         if( !result ) {
            log << coral::Error
                << "Problems encountered when removing collection '" << srcinfo.name(i)
                << "', type: " <<  srcinfo.type(i)
                << coral::MessageStream::endmsg;
         }
      }

      log << coral::Info << "Collection(s) deleted" <<  coral::MessageStream::endmsg;
      return 0;

   }
   catch( pool::Exception& poolException )
   {
      std::cerr << "pool::Exception: " << poolException.what() << std::endl;
      return 1;
   }
   catch( std::exception& exception )
   {
      std::cerr << "std::Exception: " << exception.what() << std::endl;
      return 1;
   }
   catch( ... )
   {
      std::cerr << "Unknown exception. " << std::endl;
      return 1;
   }
}





